import ast
import pathlib
from typing import List, Tuple

from flake8_stash.plugin import Plugin


def test_model_forms():
    errors = get_results("sample_model_forms.py")
    expected = [
        (15, 8, "STA011"),
        (25, 8, "STA011"),
        (33, 8, "STA012"),
        (39, 8, "STA011"),
        (49, 8, "STA042"),
    ]
    assert errors == expected


def test_model_serializers():
    errors = get_results("sample_model_serializers.py")
    expected = [
        (16, 8, "STA021"),
        (26, 8, "STA021"),
        (34, 8, "STA022"),
        (40, 8, "STA021"),
        (52, 8, "STA041"),
    ]
    assert errors == expected


def test_filtersets():
    errors = get_results("sample_filtersets.py")
    expected = [
        (16, 8, "STA031"),
        (26, 8, "STA031"),
        (34, 8, "STA032"),
        (40, 8, "STA031"),
    ]
    assert errors == expected


def get_results(filename: str) -> List[Tuple[int, int, str]]:
    current_dir = pathlib.Path(__file__).parent.resolve()
    with open(current_dir / filename, encoding="utf8") as f:
        contents = f.read()

    tree = ast.parse(contents)
    plugin = Plugin(tree)
    return [
        # retrieve position of the error and the code only
        (lineno, col_offset, msg.split()[0][:-1])
        for lineno, col_offset, msg, _ in plugin.run()
    ]



from django.contrib.auth.models import User
from django_filters import constants, rest_framework as filters
from django_filters.constants import ALL_FIELDS
from django_filters.rest_framework import FilterSet


class Mixin:
    pass


class Test1FilterSet(Mixin, filters.FilterSet):
    test_field = filters.CharFilter()

    class Meta:
        model = User
        fields = ALL_FIELDS


class Test2FilterSet(FilterSet):
    test_field = filters.CharFilter()

    class Meta:
        """"""

        model = User
        fields = constants.ALL_FIELDS


class Test3FilterSet(FilterSet):
    test_field = filters.CharFilter()

    class Meta:
        model = User
        exclude = ("password",)


class Test4FilterSet(filters.FilterSet):
    class Meta:
        model = User
        fields = "__all__"


class Test5FilterSet(Test1FilterSet):
    class Meta:
        model = User
        fields = ("first_name", "last_name")


class Test6FilterSet(Test1FilterSet):
    pass



import ast

from .dunder_all import get_meta_class, inspect_class


class Plugin:
    name = "flake8-stash"
    version = "0.11.0"

    def __init__(self, tree):
        self.tree = tree

    def run(self):
        classes_with_meta = [
            node
            for node in ast.walk(self.tree)
            if isinstance(node, ast.ClassDef) and get_meta_class(node)
        ]
        for class_def in classes_with_meta:
            node, issue = inspect_class(class_def)
            if node and issue:
                msg = f"{issue.code}: {issue.msg}"
                yield node.lineno, node.col_offset, msg, type(self)



import ast
from typing import Optional, Tuple

from .issues import Base, Issue, get_dunder_all_issue


def get_meta_class(class_def: ast.ClassDef) -> ast.ClassDef:
    for node in reversed(class_def.body):
        # this ignores the fact that class body can have arbitrary python code
        # For example, the `Meta` can be placed inside an `if` statement
        if isinstance(node, ast.ClassDef) and node.name == "Meta":
            return node


def inspect_class(
    class_def: ast.ClassDef,
) -> Tuple[Optional[ast.Assign], Optional[Issue]]:
    meta = get_meta_class(class_def)
    if all_fields_node := get_all_fields_node(meta):
        base = get_base(class_def)
        return all_fields_node, get_dunder_all_issue("fields", base)
    elif exclude_node := get_exclude_node(meta):
        base = get_base(class_def)
        return exclude_node, get_dunder_all_issue("exclude", base)
    else:
        return None, None


def get_base(class_def: ast.ClassDef) -> Base:
    for base in ["ModelForm", "ModelSerializer", "FilterSet"]:
        if is_subclass_of(class_def, base):
            return base


def is_subclass_of(class_def: ast.ClassDef, parent: str) -> bool:
    return any(
        (isinstance(base, ast.Name) and base.id == parent)
        or (isinstance(base, ast.Attribute) and base.attr == parent)
        for base in class_def.bases
    )


def get_exclude_node(meta: ast.ClassDef) -> Optional[ast.Assign]:
    """
    check if the Meta class defines an exclude attribute.
    Ignoring that Meta can have arbitrary python code i.e. exclude can
    be nested under an if statement
    """
    assign_stmts = [node for node in meta.body if isinstance(node, ast.Assign)]
    exclude_nodes = [
        target
        for assign_stmt in assign_stmts
        for target in assign_stmt.targets
        if getattr(target, "id", None) == "exclude"
    ]
    # return the last one since that's the one in effect
    return exclude_nodes[-1] if exclude_nodes else None


def get_all_fields_node(meta: ast.ClassDef) -> Optional[ast.Assign]:
    """
    check if the Meta class defines fields attribute and its value is either
    the string "__all__" or the python variable ALL_FIELDS.
    Ignoring that Meta can have arbitrary python code i.e. fields can
    be nested under an if statement
    """
    fields_stmts = [
        node
        for node in meta.body
        if isinstance(node, ast.Assign)
        and any(getattr(target, "id", None) == "fields" for target in node.targets)
    ]
    # return the last one since that's the one in effect
    for fields_stmt in reversed(fields_stmts):
        is_dunder_all_str = (
            isinstance(fields_stmt.value, ast.Constant)
            and fields_stmt.value.value == "__all__"
        )
        is_dunder_all_var = (
            isinstance(fields_stmt.value, ast.Name)
            and fields_stmt.value.id == "ALL_FIELDS"
            or isinstance(fields_stmt.value, ast.Attribute)
            and fields_stmt.value.attr == "ALL_FIELDS"
        )
        if is_dunder_all_str or is_dunder_all_var:
            return fields_stmt



import rest_framework
from django.contrib.auth.models import User
from rest_framework import serializers
from rest_framework.serializers import ALL_FIELDS, ModelSerializer


class Mixin:
    pass


class Test1Serializer(Mixin, rest_framework.serializers.ModelSerializer):
    test_field = serializers.CharField()

    class Meta:
        model = User
        fields = ALL_FIELDS


class Test2Serializer(ModelSerializer):
    test_field = serializers.CharField()

    class Meta:
        """"""

        model = User
        fields = serializers.ALL_FIELDS


class Test3Serializer(ModelSerializer):
    test_field = serializers.CharField()

    class Meta:
        model = User
        exclude = ("password",)


class Test4Serializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = "__all__"


class Test5Serializer(Test1Serializer):
    class Meta:
        model = User
        fields = ("first_name", "last_name")


class Test6Serializer(Test1Serializer):
    class Meta:
        model = User
        fields = "__all__"



from django import forms
from django.contrib.auth.models import User
from django.forms import ALL_FIELDS, ModelForm


class Mixin:
    pass


class Test1Form(Mixin, forms.models.ModelForm):
    test_field = forms.CharField()

    class Meta:
        model = User
        fields = ALL_FIELDS


class Test2Form(ModelForm):
    test_field = forms.CharField()

    class Meta:
        """"""

        model = User
        fields = forms.ALL_FIELDS


class Test3Form(ModelForm):
    test_field = forms.CharField()

    class Meta:
        model = User
        exclude = ("password",)


class Test4Form(forms.ModelForm):
    class Meta:
        model = User
        fields = "__all__"


class Test5Form(Test1Form):
    pass


class Test6Form(Test1Form):
    class Meta:
        model = User
        exclude = ("first_name",)



import pathlib
from argparse import ArgumentParser
from datetime import date

CHANGELOG_RELATIVE_PATH = "CHANGELOG.md"
CURRENT_DIR = pathlib.Path(__file__).parent.resolve()
CHANGELOG_ABS_PATH = CURRENT_DIR.parent / CHANGELOG_RELATIVE_PATH
UNRELEASED_LINE = "## [UNRELEASED]"
NEW_VERSION_LINE_TEMPLATE = "## [{version}] - {date}\n"


def main(version):
    lines = get_changelog_file_content()
    unreleased_line_number = find_unreleased_line_number(lines)
    lines = add_new_version_to_changelog(lines, unreleased_line_number, version)
    update_changelog_file(lines)


def get_changelog_file_content():
    with open(CHANGELOG_ABS_PATH, encoding="utf-8") as f:
        return f.readlines()


def find_unreleased_line_number(lines):
    unreleased_line_number = None
    for i, line in enumerate(lines):
        if line.strip().lower() == UNRELEASED_LINE.lower():
            unreleased_line_number = i
            break

    if unreleased_line_number is None:
        # Abort the release process as we're not able to update the changelog.
        # If, for some reason, there is no need to update the changelog, comment
        # the step to update the changelog in pyproject.toml
        raise ChangelogUpdateError(
            f"Unable to find a line with the text '{UNRELEASED_LINE}' when "
            f"trying to update the changelog at '{CHANGELOG_RELATIVE_PATH}'."
        )
    return unreleased_line_number


def add_new_version_to_changelog(lines, unreleased_line_number, version):
    new_version_line = NEW_VERSION_LINE_TEMPLATE.format(
        version=version, date=date.today()
    )
    return (
        lines[: unreleased_line_number + 1]
        + ["\n", new_version_line]
        + lines[unreleased_line_number + 1 :]
    )


def update_changelog_file(lines):
    with open(CHANGELOG_ABS_PATH, "w", encoding="utf-8") as f:
        f.writelines(lines)


class ChangelogUpdateError(Exception):
    """Unable to update the changelog with the new version"""


if __name__ == "__main__":
    parser = ArgumentParser(
        description=(
            "Replace UNRELEASED with the new version and current date and add "
            "a new unreleased section to the changelog."
        )
    )
    parser.add_argument(
        "-n",
        "--new-version",
        dest="version",
        help="The new version to be released",
        metavar="NEW_VERSION",
    )
    args = parser.parse_args()

    main(args.version)



from dataclasses import dataclass
from typing import Literal, Optional

Attr = Literal["fields", "exclude"]
Base = Optional[Literal["ModelForm", "ModelSerializer", "FilterSet"]]


def get_dunder_all_issue(attr: Attr, base: Base = None) -> "Issue":
    mapping = {
        ("fields", "ModelForm"): STA011,
        ("exclude", "ModelForm"): STA012,
        ("fields", "ModelSerializer"): STA021,
        ("exclude", "ModelSerializer"): STA022,
        ("fields", "FilterSet"): STA031,
        ("exclude", "FilterSet"): STA032,
        ("fields", None): STA041,
        ("exclude", None): STA042,
    }
    return mapping[(attr, base)]


@dataclass
class Issue:
    code: str
    msg: str
    attr: Attr
    base: Base = None


STA011 = Issue(
    code="STA011",
    msg="The use of `fields = '__all__'` in model forms is not allowed. List the fields one by one instead.",
    attr="fields",
    base="ModelForm",
)
STA012 = Issue(
    code="STA012",
    msg="The use of `exclude` in model forms is not allowed. Use `fields` instead.",
    attr="exclude",
    base="ModelForm",
)
STA021 = Issue(
    code="STA021",
    msg="The use of `fields = '__all__'` in model serializers is not allowed. List the fields one by one instead.",
    attr="fields",
    base="ModelSerializer",
)
STA022 = Issue(
    code="STA022",
    msg="The use of `exclude` in model serializers is not allowed. Use `fields` instead.",
    attr="exclude",
    base="ModelSerializer",
)
STA031 = Issue(
    code="STA031",
    msg="The use of `fields = '__all__'` in filtersets is not allowed. List the fields one by one instead.",
    attr="fields",
    base="FilterSet",
)
STA032 = Issue(
    code="STA032",
    msg="The use of `exclude` in filtersets is not allowed. Use `fields` instead.",
    attr="exclude",
    base="FilterSet",
)
STA041 = Issue(
    code="STA041",
    msg="The use of `fields = '__all__'` is not allowed. List the fields one by one instead.",
    attr="fields",
)
STA042 = Issue(
    code="STA042",
    msg="The use of `exclude` is not allowed. Use `fields` instead.",
    attr="exclude",
)
